@@ -24,6 +24,8 @@
 #include <ctype.h>
 #include <string.h>
 #include <sys/wait.h>
+#include <errno.h>
+
 #include <glib.h>
 #include <gio/gio.h>
 
@@ -34,7 +36,7 @@
 
 enum {
         AUTOSTART_LAUNCH_SPAWN = 0,
-        AUTOSTART_LAUNCH_ACTIVATE,
+        AUTOSTART_LAUNCH_ACTIVATE
 };
 
 enum {
@@ -42,7 +44,7 @@ enum {
         GSM_CONDITION_IF_EXISTS     = 1,
         GSM_CONDITION_UNLESS_EXISTS = 2,
         GSM_CONDITION_GNOME         = 3,
-        GSM_CONDITION_UNKNOWN       = 4,
+        GSM_CONDITION_UNKNOWN       = 4
 };
 
 #define GSM_SESSION_CLIENT_DBUS_INTERFACE "org.gnome.SessionClient"
@@ -77,7 +79,7 @@ enum {
 
 enum {
         PROP_0,
-        PROP_DESKTOP_FILENAME,
+        PROP_DESKTOP_FILENAME
 };
 
 static guint signals[LAST_SIGNAL] = { 0 };
@@ -100,17 +102,9 @@ static gboolean
 is_disabled (GsmApp *app)
 {
         GsmAutostartAppPrivate *priv;
-        gboolean                autorestart = FALSE;
 
         priv = GSM_AUTOSTART_APP (app)->priv;
 
-        if (egg_desktop_file_has_key (priv->desktop_file,
-                                      "X-GNOME-AutoRestart", NULL)) {
-                autorestart = egg_desktop_file_get_boolean (priv->desktop_file,
-                                                            "X-GNOME-AutoRestart",
-                                                            NULL);
-        }
-
         /* X-GNOME-Autostart-enabled key, used by old gnome-session */
         if (egg_desktop_file_has_key (priv->desktop_file,
                                       "X-GNOME-Autostart-enabled", NULL) &&
@@ -307,6 +301,7 @@ setup_condition_monitor (GsmAutostartApp *app)
         key = NULL;
         res = parse_condition_string (app->priv->condition_string, &kind, &key);
         if (! res) {
+                g_free (key);
                 return;
         }
 
@@ -373,6 +368,8 @@ setup_condition_monitor (GsmAutostartApp *app)
                 disabled = TRUE;
         }
 
+        g_free (key);
+
         /* FIXME: cache the disabled value? */
 }
 
@@ -465,8 +462,6 @@ gsm_autostart_app_set_desktop_filename (GsmAutostartApp *app,
 {
         GError *error;
 
-        g_debug ("GsmAutostartApp: setting desktop filename to %s", desktop_filename);
-
         if (app->priv->desktop_file != NULL) {
                 egg_desktop_file_free (app->priv->desktop_file);
                 app->priv->desktop_file = NULL;
@@ -579,6 +574,8 @@ gsm_autostart_app_dispose (GObject *object)
         if (priv->condition_monitor) {
                 g_file_monitor_cancel (priv->condition_monitor);
         }
+
+        G_OBJECT_CLASS (gsm_autostart_app_parent_class)->dispose (object);
 }
 
 static gboolean
@@ -617,6 +614,7 @@ is_conditionally_disabled (GsmApp *app)
         key = NULL;
         res = parse_condition_string (priv->condition_string, &kind, &key);
         if (! res) {
+                g_free (key);
                 return TRUE;
         }
 
@@ -649,10 +647,7 @@ is_conditionally_disabled (GsmApp *app)
         /* Set initial condition */
         priv->condition = !disabled;
 
-        if (disabled) {
-                g_debug ("app %s is disabled by AutostartCondition",
-                         gsm_app_peek_id (app));
-        }
+        g_free (key);
 
         return disabled;
 }
@@ -682,10 +677,57 @@ app_exited (GPid             pid,
         }
 }
 
+static int
+_signal_pid (int pid,
+             int signal)
+{
+        int status = -1;
+
+        /* perhaps block sigchld */
+        g_debug ("GsmAutostartApp: sending signal %d to process %d", signal, pid);
+        errno = 0;
+        status = kill (pid, signal);
+
+        if (status < 0) {
+                if (errno == ESRCH) {
+                        g_warning ("Child process %d was already dead.",
+                                   (int)pid);
+                } else {
+                        g_warning ("Couldn't kill child process %d: %s",
+                                   pid,
+                                   g_strerror (errno));
+                }
+        }
+
+        /* perhaps unblock sigchld */
+
+        return status;
+}
+
 static gboolean
 autostart_app_stop_spawn (GsmAutostartApp *app,
                           GError         **error)
 {
+        int res;
+
+        if (app->priv->pid < 1) {
+                g_set_error (error,
+                             GSM_APP_ERROR,
+                             GSM_APP_ERROR_STOP,
+                             "Not running");
+                return FALSE;
+        }
+
+        res = _signal_pid (app->priv->pid, SIGTERM);
+        if (res != 0) {
+                g_set_error (error,
+                             GSM_APP_ERROR,
+                             GSM_APP_ERROR_STOP,
+                             "Unable to stop: %s",
+                             g_strerror (errno));
+                return FALSE;
+        }
+
         return TRUE;
 }
 
@@ -737,6 +779,7 @@ autostart_app_start_spawn (GsmAutostartApp *app,
 
         env[0] = g_strdup_printf ("DESKTOP_AUTOSTART_ID=%s", startup_id);
 
+        local_error = NULL;
         command = egg_desktop_file_parse_exec (app->priv->desktop_file,
                                                NULL,
                                                &local_error);
@@ -750,6 +793,7 @@ autostart_app_start_spawn (GsmAutostartApp *app,
         g_debug ("GsmAutostartApp: starting %s: command=%s startup-id=%s", app->priv->desktop_id, command, startup_id);
         g_free (command);
 
+        g_free (app->priv->startup_id);
         local_error = NULL;
         success = egg_desktop_file_launch (app->priv->desktop_file,
                                            NULL,
@@ -900,11 +944,12 @@ gsm_autostart_app_restart (GsmApp  *app,
         GError  *local_error;
         gboolean res;
 
+        /* ignore stop errors - it is fine if it is already stopped */
         local_error = NULL;
         res = gsm_app_stop (app, &local_error);
         if (! res) {
-                g_propagate_error (error, local_error);
-                return FALSE;
+                g_debug ("GsmAutostartApp: Couldn't stop app: %s", local_error->message);
+                g_error_free (local_error);
         }
 
         res = gsm_app_start (app, &local_error);
